#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ioc.component

class Extension(ioc.component.Extension):
    def load(self, config, container_builder):
        definition = ioc.component.Definition('ioc.event.Dispatcher', kwargs={
            'logger': ioc.component.Reference('logger')
        })
        container_builder.add('ioc.extra.event_dispatcher', definition)

    def post_build(self, container_builder, container):
        dispatcher = container.get('ioc.extra.event_dispatcher')

        for id in container_builder.get_ids_by_tag('event.listener'):
            definition = container_builder.get(id)
            for option in definition.get_tag('event.listener'):
                if 'name' not in option:
                    break

                if 'method' not in option:
                    break

                if 'priority' not in option:
                    option['priority'] = 0

                if type(option['method']) == str:
                    method = getattr(container.get(id), option['method'])
                else:
                    method = option['method']

                dispatcher.add_listener(option['name'], method, option['priority'])



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
    Resource locator

    This code is based on the jinja2.loaders module
"""
from os import path

class ResourceNotFound(Exception):
    pass

def split_resource_path(resource: str) -> list[str]:
    """Split a path into segments and perform a sanity check.  If it detects
    '..' in the path it will raise a `TemplateNotFound` error.
    """
    pieces = []
    for piece in resource.split('/'):
        if path.sep in piece \
           or (path.altsep and path.altsep in piece) or \
           piece == path.pardir:
            raise ResourceNotFound(resource)
        elif piece and piece != '.':
            pieces.append(piece)
    return pieces


class BaseLocator(object):
    def locate(self, resource: str) -> str:
        raise ResourceNotFound(resource)


class FileSystemLocator(BaseLocator):
    """Locale a ressource from the file system.  This Locator can find ressources
    in folders on the file system and is the preferred way to load them.

    The Locator takes the path to the resource as string, or if multiple
    locations are wanted a list of them which is then looked up in the
    given order:

    >>> locator = FileSystemLocator('/path/to/templates')
    >>> locator = FileSystemLocator(['/path/to/templates', '/other/path'])

    """

    def __init__(self, searchpath):
        if isinstance(searchpath, str):
            searchpath = [searchpath]

        self.searchpath = list(searchpath)

    def locate(self, resource):
        pieces = split_resource_path(resource)
        for searchpath in self.searchpath:
            filename = path.join(searchpath, *pieces)

            if not path.exists(filename):
                continue

            return filename

        raise ResourceNotFound(resource)


class PackageLocator(BaseLocator):
    """Load ressource from python eggs or packages.  It is constructed with
    the name of the python package and the path to the resource folder in that
    package::

        locator = PackageLocator('mypackage', 'views')

    If the package path is not given, ``'resources'`` is assumed.
    """

    def __init__(self, package_name: str, package_path: str = 'resources') -> None:
        try:
            # Python 3.9+
            from importlib import resources
            self.resources = resources
        except ImportError:
            # Fallback for older Python versions
            import importlib_resources as resources
            self.resources = resources
        
        self.package_name = package_name
        self.package_path = package_path

    def locate(self, resource: str) -> str:
        pieces = split_resource_path(resource)
        
        try:
            # Try to access the resource
            package = self.resources.files(self.package_name)
            if self.package_path:
                package = package / self.package_path
            
            for piece in pieces:
                package = package / piece
            
            if not package.is_file():
                raise ResourceNotFound(resource)
            
            # For Python 3.9+, we need to handle the path properly
            if hasattr(package, '__fspath__'):
                return str(package)
            else:
                # Use as_file context manager for temporary access
                with self.resources.as_file(package) as path:
                    return str(path)
        except (AttributeError, FileNotFoundError, ModuleNotFoundError):
            raise ResourceNotFound(resource)

class FunctionLocator(BaseLocator):
    """A locator that is passed a function which does the searching.  The
    function becomes the name of the resource passed and has to return
    a string with the resource sourc.

    >>> def load_resource(name):
    ...     if name == 'index.html':
    ...         return '...'
    ...
    >>> locator = FunctionLocator(load_resource)
    """

    def __init__(self, load_func):
        self.load_func = load_func

    def locate(self, resource):
        rv = self.load_func(resource)
        if rv is None:
            raise ResourceNotFound(resource)

        return rv

class PrefixLocator(BaseLocator):
    """A locator that is passed a dict of locators where each Locator is bound
    to a prefix.  The prefix is delimited from the resource by a slash per
    default, which can be changed by setting the `delimiter` argument to
    something else::

        Locator = PrefixLocator({
            'app1':     PackageLocator('mypackage.app1'),
            'app2':     PackageLocator('mypackage.app2')
        })

    By loading ``'app1/index.html'`` the file from the app1 package is loaded,
    by loading ``'app2/index.html'`` the file from the second.
    """

    def __init__(self, mapping, delimiter='/'):
        self.mapping = mapping
        self.delimiter = delimiter

    def locate(self, resource):
        try:
            prefix, name = resource.split(self.delimiter, 1)
            locator = self.mapping[prefix]
        except (ValueError, KeyError):
            raise ResourceNotFound(resource)

        try:
            return locator.locate(name)
        except ResourceNotFound:
            # re-raise the exception with the correct fileame here.
            # (the one that includes the prefix)
            raise ResourceNotFound(resource)

class ChoiceLocator(BaseLocator):
    """This locator works like the `PrefixLocator` just that no prefix is
    specified.  If a resource could not be found by one locator the next one
    is tried.

    >>> locator = ChoiceLocator([
    ...     FileSystemLocator('/path/to/user/templates'),
    ...     FileSystemLocator('/path/to/system/templates')
    ... ])

    This is useful if you want to allow users to override builtin resources
    from a different location.
    """
    def __init__(self, locators):
        self.locators = locators

    def locate(self, resource):
        for locator in self.locators:
            try:
                return locator.locate(resource)
            except ResourceNotFound:
                pass

        raise ResourceNotFound(resource)



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ioc.loader, ioc.component, ioc.exceptions
import os

class Extension(ioc.component.Extension):
    def load(self, config, container_builder):

        path = os.path.dirname(os.path.abspath(__file__))

        loader = ioc.loader.YamlLoader()
        loader.load("%s/resources/config/redis.yml" % path, container_builder)

        container_builder.parameters.set('ioc.extra.redis.default_connection', config.get('default', 'default'))

        self.configure_connections(config, container_builder)

    def configure_connections(self, config, container_builder):

        manager = container_builder.get('ioc.extra.redis')

        for name, parameters in config.get_dict('connections', {}).iteritems():
            id = "ioc.extra.redis.connection.%s" % name

            container_builder.add(id, ioc.component.Definition('redis.ConnectionPool', kwargs={
                'host':             parameters.get('host', 'localhost'),
                'port':             parameters.get_int('port', 6379),
                'db':               parameters.get('db', 0), 
                'password':         parameters.get('password', None),
                'socket_timeout':   parameters.get('socket_timeout', None), 
                'encoding':         parameters.get('encoding', 'utf-8'),
                'encoding_errors':  parameters.get('encoding_errors', 'strict'), 
                'decode_responses': parameters.get('decode_responses', False),
            }))

            manager.add_call('add_connection', arguments=[name, ioc.component.Reference(id)])

        for name, parameters in config.get_dict('clients', {}).iteritems():
            id_connection = "ioc.extra.redis.connection.%s" % parameters.get('connection')
            id = "ioc.extra.redis.client.%s" % name

            if not container_builder.has(id_connection):
                raise ioc.exceptions.UnknownService("Redis client defined an undefined service: %s" % id_connection)

            container_builder.add(id, ioc.component.Definition('redis.StrictRedis', kwargs={
                'connection_pool':  ioc.component.Reference(id_connection),
            }))

            manager.add_call('add_client', arguments=[name, ioc.component.Reference(id)])



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ioc.extra.command import Command

class StartCommand(Command):

    def __init__(self, flask):
        self.flask = flask

    def initialize(self, parser):
        parser.description = 'Start flask integrated server'
        parser.add_argument('--host', default='0.0.0.0', help="the host to bind the port")
        parser.add_argument('--port', default=5000, type=int, help="the port to listen")

    def execute(self, args, output):
        output.write("Starting flask...\n")

        options = {
            'host': args.host,
            'port': args.port,
        }

        # flask autoreload cannot read from arguments line
        if args.debug:
            options['debug'] = True
            options['use_reloader'] = False
            options['use_debugger'] = False

        self.flask.run(**options)



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

class Manager(object):
    def __init__(self, default=None, connections=None, clients=None):
        self.connections = connections or {}
        self.clients = clients or {}
        self.default = default or False

    def add_connection(self, name, connection):
        self.connections[name] = connection

    def get_connection(self, name=None):
        if len(self.connections) == 1:
            return self.connections.values()[0]

        if name in self.connections:
            return self.connections[name]

        raise KeyError('Unable to find the the valid connection')

    def add_client(self, name, client):
        self.clients[name] = client

    def get_default_client(self):
        if len(self.clients) == 1:
            return self.clients.values()[0]

        if not self.default:
            raise KeyError('No default client set')

        return self.get_client(self, name=self.default)
        
    def get_client(self, name=None):
        if not name:
            return self.get_default_client()

        if name in self.clients:
            return self.clients[name]

        raise KeyError('Unable to find the the valid connection')



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ioc.component import Container

class Proxy(object):
    __slots__ = ["_obj", "_container", "_id", "__weakref__"]

    def __init__(self, container: 'Container', id: str) -> None:
        object.__setattr__(self, "_obj", None)
        object.__setattr__(self, "_container", container)
        object.__setattr__(self, "_id", id)
        
    #
    # proxying (special cases)
    #
    def __getattribute__(self, name: str) -> Any:
        build_object(object, self)
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __delattr__(self, name: str) -> None:
        build_object(object, self)
        delattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name: str, value: Any) -> None:
        build_object(object, self)
        setattr(object.__getattribute__(self, "_obj"), name, value)
    
    def __nonzero__(self) -> bool:
        build_object(object, self)
        return bool(object.__getattribute__(self, "_obj"))

    def __str__(self) -> str:
        build_object(object, self)
        return "<Proxy " + str(object.__getattribute__(self, "_obj")) + ">"

    def __repr__(self) -> str:
        build_object(object, self)
        return repr(object.__getattribute__(self, "_obj"))

def build_object(object: type, instance: Proxy) -> None:
    if object.__getattribute__(instance, "_obj") is None:
        container = object.__getattribute__(instance, "_container")
        id = object.__getattribute__(instance, "_id")

        object.__setattr__(instance, "_obj", container.get(id))



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
    The default yaml loader does not respect the ordering key while loading a dictionay.
    This can lead to error while ordering is defining key by key

    reference: https://gist.github.com/enaeseth/844388

"""

from typing import Any, Iterator, Optional, Union
import yaml
import yaml.constructor

from collections import OrderedDict

def deepcopy(value: Any) -> Any:
    """
    The default copy.deepcopy seems to copy all objects and some are not
    `copy-able`.

    We only need to make sure the provided data is a copy per key, object does
    not need to be copied.
    """

    if not isinstance(value, (dict, list, tuple)):
        return value

    if isinstance(value, dict):
        copy = {}
        for k, v in value.items():
            copy[k] = deepcopy(v)
        return copy
            
    if isinstance(value, tuple):
        copy = list(range(len(value)))

        for k in get_keys(list(value)):
            copy[k] = deepcopy(value[k])

        return tuple(copy)

    if isinstance(value, list):
        copy = list(range(len(value)))

        for k in get_keys(value):
            copy[k] = deepcopy(value[k])

        return copy

    return value

def is_scalar(value: Any) -> bool:
    return isinstance(value, (str))

def is_iterable(value: Any) -> bool:
    return isinstance(value, (dict, list))

def get_keys(arguments: Union[list[Any], dict[str, Any]]) -> Union[range, Any, list[Any]]:
    if isinstance(arguments, (list)):
        return range(len(arguments))

    if isinstance(arguments, (dict)):
        return arguments.keys()

    return []

class Dict(object):
    def __init__(self, data: Optional[dict[str, Any]] = None) -> None:
        self.data = data or {}

    def get(self, name: str, default: Optional[Any] = None) -> Any:
        data = self.data
        for name in name.split("."):
            if name in data:
                data = data[name]
            else:
                return default

        return data

    def get_dict(self, name: str, default: Optional[dict[str, Any]] = None) -> 'Dict':
        default = default or {}
        value = self.get(name, default)

        if not isinstance(value, Dict):
            value = Dict(value)

        return value

    def get_int(self, name: str, default: Optional[int] = None) -> int:
        return int(self.get(name, default))

    def get_all(self, name: str, default: Optional[dict[str, Any]] = None) -> dict[str, Any]:
        return self.get_dict(name, default).all()

    def all(self) -> dict[str, Any]:
        def walk(data: Any) -> Any:
            all = {}

            if not isinstance(data, dict):
                return data

            for v, d in data.items():
                if isinstance(d, Dict):
                    all[v] = d.all()
                else:
                    all[v] = d

                if is_iterable(all[v]):
                    walk(all[v])

            return all

        return walk(self.data)

    def iteritems(self) -> Iterator[Any]:
        return self.data.items()

    def __iter__(self) -> Iterator[Any]:
        return iter(self.data)

    def __getitem__(self, key: str) -> Any:
        return self.data[key]

class OrderedDictYAMLLoader(yaml.Loader):
    """
    A YAML loader that loads mappings into ordered dictionaries.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        yaml.Loader.__init__(self, *args, **kwargs)

        self.add_constructor('tag:yaml.org,2002:map', type(self).construct_yaml_map)
        self.add_constructor('tag:yaml.org,2002:omap', type(self).construct_yaml_map)

    def construct_yaml_map(self, node: Any) -> Any:
        data = OrderedDict()
        yield data
        value = self.construct_mapping(node)
        data.update(value)

    def construct_mapping(self, node: Any, deep: bool = False) -> OrderedDict:
        if isinstance(node, yaml.MappingNode):
            self.flatten_mapping(node)
        else:
            raise yaml.constructor.ConstructorError(None, None,
                                                    'expected a mapping node, but found %s' % node.id,
                                                    node.start_mark)

        mapping = OrderedDict()
        for key_node, value_node in node.value:
            # key = self.construct_object(key_node, deep=deep)

            key = key_node.value
            try:
                hash(key)
            except TypeError as exc:
                raise yaml.constructor.ConstructorError('while constructing a mapping',
                                                        node.start_mark,
                                                        'found unacceptable key (%s)' % exc,
                                                        key_node.start_mark)
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value

        return mapping



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ioc.loader, ioc.component, ioc.exceptions
from ioc.component import Definition

class Extension(ioc.component.Extension):
    def load(self, config, container_builder):
        extensions = container_builder.parameters.get('ioc.extensions')

        locator_map = {}

        for extension in extensions:
            arguments = [[
                Definition('ioc.locator.FileSystemLocator', arguments=["%s/resources/%s" % (container_builder.parameters.get('project.root_folder'), extension)]),
                Definition('ioc.locator.PackageLocator', arguments=[extension], kwargs={'package_path': 'resources'}),
                Definition('ioc.locator.PackageLocator', arguments=[extension], kwargs={'package_path': ''})
            ]]
            
            locator_map[extension] = Definition('ioc.locator.ChoiceLocator', arguments=arguments)

        container_builder.add('ioc.locator', Definition('ioc.locator.PrefixLocator', arguments=[locator_map], kwargs={'delimiter': ':'}))



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Optional, Union
from .exceptions import UnknownService, ParameterHolderIsFrozen, UnknownParameter, RecursiveParameterResolutionError, AbstractDefinitionInitialization, CyclicReference
from .proxy import Proxy
from .misc import deepcopy, get_keys, is_iterable

import importlib, inspect, re, logging

HAS_PYDANTIC = False
try:
    from pydantic import BaseModel as PydanticBaseModel
    HAS_PYDANTIC = True
except ImportError:
    pass

class Reference(object):
    def __init__(self, id: str, method: Optional[str] = None) -> None:
        self.id = id
        self.method = method

class WeakReference(Reference):
    pass

class Definition(object):
    def __init__(self, clazz: Optional[Union[str, list[str]]] = None, arguments: Optional[list[Any]] = None, kwargs: Optional[dict[str, Any]] = None, abstract: bool = False) -> None:
        self.clazz = clazz
        self.arguments = arguments or [] 
        self.kwargs = kwargs or {}
        self.method_calls: list[list[Any]] = []
        self.property_calls: list[Any] = []
        self.tags: dict[str, list[dict[str, Any]]] = {}
        self.abstract = abstract

    def add_call(self, method: str, arguments: Optional[list[Any]] = None, kwargs: Optional[dict[str, Any]] = None) -> None:
        self.method_calls.append([
            method,
            arguments or [],
            kwargs or {}
        ])

    def add_tag(self, name: str, options: Optional[dict[str, Any]] = None) -> None:
        if name not in self.tags:
            self.tags[name] = []

        self.tags[name].append(options or {})

    def has_tag(self, name: str) -> bool:
        return name in self.tags

    def get_tag(self, name: str) -> list[dict[str, Any]]:
        if not self.has_tag(name):
            return []

        return self.tags[name]

class ParameterHolder(object):
    def __init__(self, parameters: Optional[dict[str, Any]] = None) -> None:
        self._parameters = parameters or {}
        self._frozen = False

    def set(self, key: str, value: Any) -> None:
        if self._frozen:
            raise ParameterHolderIsFrozen(key)

        self._parameters[key] = value

    def get(self, key: str) -> Any:
        if key in self._parameters:
            return self._parameters[key]

        raise UnknownParameter(key)

    def remove(self, key: str) -> None:
        del self._parameters[key]

    def has(self, key: str) -> bool:
        return key in self._parameters

    def all(self) -> dict[str, Any]:
        return self._parameters

    def __setitem__(self, key: str, value: Any) -> None:
        self.set(key, value)

    def freeze(self) -> None:
        self._frozen = True

    def is_frozen(self) -> bool:
        return self._frozen == True

class ParameterResolver(object):
    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self.re = re.compile(r"%%|%([^%\s]+)%")
        self.logger = logger
        self.stack: list[str] = []

    def _resolve(self, parameter: Any, parameter_holder: ParameterHolder) -> Any:
        if HAS_PYDANTIC and isinstance(parameter, PydanticBaseModel):
            # If the parameter is a Pydantic model, resolve its fields
            def walk_and_modify(model):
                for field in model.__fields__:
                    value = getattr(model, field)
                    if isinstance(value, PydanticBaseModel):
                        parameter = walk_and_modify(value)  # Recursive call for nested models
                    else:
                        parameter = self.resolve(value, parameter_holder)
                    
                    setattr(model, field, parameter)
                    
                    return model
                
            return walk_and_modify(parameter)
        
        if isinstance(parameter, (tuple)):
            parameter = list(parameter)
            for key in get_keys(parameter):
                parameter[key] = self.resolve(parameter[key], parameter_holder)

            return tuple(parameter)

        if is_iterable(parameter):
            for key in get_keys(parameter):
                parameter[key] = self.resolve(parameter[key], parameter_holder)

            return parameter

        if not type(parameter) == str:
            return parameter

        if parameter[0:1] == '%' and parameter[-1] == '%' and parameter_holder.has(parameter[1:-1]):
            # if self.logger:
            #     self.logger.debug("   >> Match parameter: %s" % parameter[1:-1])

            return self.resolve(parameter_holder.get(parameter[1:-1]), parameter_holder)

        def replace(matchobj):
            if matchobj.group(0) == '%%':
                return '%'

            return self.resolve(parameter_holder.get(matchobj.group(1)), parameter_holder)

        # if self.logger:
        #     self.logger.debug("   >> Start resolving parameter: %s" % parameter)

        parameter, nums = re.subn(self.re, replace, parameter)

        # print parameter
        return parameter

    def resolve(self, parameter: Any, parameter_holder: ParameterHolder) -> Any:
        if parameter in self.stack:
            raise RecursiveParameterResolutionError(" -> ".join(self.stack) + " -> " + parameter)

        parameter = deepcopy(parameter)

        self.stack.append(parameter)
        value = self._resolve(parameter, parameter_holder)
        self.stack.pop()

        return value

class Container(object):
    def __init__(self) -> None:
        self.services: dict[str, Any] = {}
        self.parameters = ParameterHolder()
        self.stack: list[str] = []

    def has(self, id: str) -> bool:
        return id in self.services

    def add(self, id: str, service: Any) -> None:
        self.services[id] = service

    def get(self, id: str) -> Any:
        if id not in self.services:
            raise UnknownService(id)

        return self.services[id]

class ContainerBuilder(Container):
    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self.services: dict[str, Definition] = {}
        self.parameters = ParameterHolder()
        self.stack: list[str] = []
        self.logger = logger
        self.parameter_resolver = ParameterResolver(logger=logger)
        self.extensions: dict[str, Any] = {}

    def add_extension(self, name: str, config: Any) -> None:
        self.extensions[name] = config

    def get_ids_by_tag(self, name: str) -> list[str]:
        return [id for id, definition in self.services.items() if definition.has_tag(name)]

    def build_container(self, container: Container) -> Container:
        if self.logger:
            self.logger.debug("Start building the container")

        extensions = []
        container.add("service_container", container)

        if not container.has('logger'):
            if not self.logger:
                import logging
                container.add("logger", logging.getLogger('app'))
            else:
                container.add("logger", self.logger)

        self.parameters.set('ioc.extensions', self.extensions.keys())

        for name, config in self.extensions.items():
            name = "%s.di.Extension" % name

            if self.logger:
                self.logger.debug("Load extension %s" % name)

            extension = self.get_class(Definition(name))()
            extension.load(config, self)

            extensions.append(extension)

        for extension in extensions:
            extension.post_load(self)

        if self.logger:
            self.logger.debug("Starting resolving all parameters!")
            
        for name, value in self.parameters.all().items():
            container.parameters.set(
                name, 
                self.parameter_resolver.resolve(value, self.parameters)
            )

        if self.logger:
            self.logger.debug("End resolving all parameters!")

        for extension in extensions:
            extension.pre_build(self, container)

        # resolve services
        for id, definition in self.services.items():
            if definition.abstract:
                continue

            self.get_service(id, definition, container)

        for extension in extensions:
            extension.post_build(self, container)

        if self.logger:
            self.logger.debug("Building container is over!")

        if container.has('ioc.extra.event_dispatcher'):
            container.get('ioc.extra.event_dispatcher').dispatch('ioc.container.built', {
                'container': container,
                'container_builder': self
            })

        for extension in extensions:
            extension.start(container)

        return container

    def create_definition(self, id: str) -> Definition:
        abstract = self.services[id]

        definition = Definition(
            clazz=abstract.clazz, 
            arguments=deepcopy(abstract.arguments),
            kwargs=deepcopy(abstract.kwargs),
            abstract=False,
        )

        definition.method_calls = deepcopy(abstract.method_calls)
        definition.property_calls = deepcopy(abstract.property_calls)
        definition.tags = deepcopy(abstract.tags)

        return definition

    def get_class(self, definition: Definition) -> Any:
        clazz = self.parameter_resolver.resolve(definition.clazz, self.parameters)

        if isinstance(clazz, list):
            module = clazz[0]
            function = clazz[1]
        else:
            module = ".".join(clazz.split(".")[0:-1])
            function = clazz.split(".")[-1]

        module = importlib.import_module(module)

        function = function.split(".")
        clazz = getattr(module, function[0])

        if len(function) == 2:
            return getattr(clazz, function[1])

        return clazz

    def get_instance(self, definition: Definition, container: Container) -> Any:

        klass = self.get_class(definition)

        if self.logger:
            self.logger.debug("Create instance for %s" % klass)

        if inspect.isclass(klass) or inspect.isfunction(klass) or inspect.ismethod(klass):
            args = self.set_services(definition.arguments, container)
            kwargs = self.set_services(definition.kwargs, container)

            instance = klass(*args, **kwargs)
        else:
            # module object ...
            instance = klass

        for call in definition.method_calls:
            method, args, kwargs = call

            if self.logger:
                self.logger.debug(" > Call method: %s on class: %s" % (method, instance))

            attr = getattr(instance, method)

            if not attr:
                # handle property definition
                setattr(instance, method, self.set_services(args, container)[0])
            else:
                attr(*self.set_services(args, container), **self.set_services(kwargs, container))
            
        if self.logger:
            self.logger.debug("End creating instance %s" % instance)

        return instance

    def get_service(self, id: str, definition: Definition, container: Container) -> Any:
        if definition.abstract:
            raise AbstractDefinitionInitialization("The ContainerBuiler try to build an abstract definition, id=%s, class=%s" % (id, definition.clazz))

        if container.has(id):
            return container.get(id)

        if id in self.stack:
            if self.logger:
                self.logger.error("ioc.exceptions.CyclicReference: " + " -> ".join(self.stack) + " -> " + id)

            raise CyclicReference(" -> ".join(self.stack) + " -> " + id)

        self.stack.append(id)
        instance = self.get_instance(definition, container)
        container.add(id, instance)
        self.stack.pop()

        return instance

    def retrieve_service(self, value: Any, container: Container) -> Any:
        if isinstance(value, (Reference, WeakReference)) and not container.has(value.id) and not self.has(value.id):
            raise UnknownService(value.id)

        if isinstance(value, (Reference)):
            if not container.has(value.id):
                service = self.get_service(value.id, self.get(value.id), container)
            else:
                service = container.get(value.id)

            # a reference can point a service's method, and not the service itself...
            if value.method:
                return getattr(service, value.method)

            return service

        if isinstance(value, (WeakReference)):
            # if the container already has the service return the service and not the proxy
            if not container.has(value.id):
                return Proxy(container, value.id)

            return container.get(value.id)

        if isinstance(value, Definition):
            return self.get_instance(value, container)

        if is_iterable(value):
            return self.set_services(value, container)

        if isinstance(value, (tuple)):
            return tuple(self.set_services(list(value), container))

        return self.parameter_resolver.resolve(value, self.parameters)

    def set_services(self, arguments: Union[list[Any], dict[str, Any]], container: Container) -> Union[list[Any], dict[str, Any]]:
        for pos in get_keys(arguments):
            arguments[pos] = self.retrieve_service(arguments[pos], container)

        return arguments

class Extension(object):
    def load(self, config: Any, container_builder: ContainerBuilder) -> None:
        pass

    def post_load(self, container_builder: ContainerBuilder) -> None:
        pass

    def pre_build(self, container_builder: ContainerBuilder, container: Container) -> None:
        pass

    def post_build(self, container_builder: ContainerBuilder, container: Container) -> None:
        pass

    def start(self, container: Container) -> None:
        pass
